      PROGRAM PCTFR
      
      implicit none
      
      include 'mpif.h'
      include '../include/allprogs.inc'
      include '../include/info.inc'
      include '../include/com_01.inc'
      include '../include/mem_time.inc'
      
      logical confine_3D
      data confine_3D / .false. / 
      
C     Local variables, used only in the main routine
      real center(2)

      integer first_iq, last_iq, max_iq_slab,
     +     first_z,last_z, first_hk,last_hk, max_hk_slab, num_hkl,
     +     den_min_x, den_max_x, den_min_y, den_max_y, 
     +     den_min_z, den_max_z, node, mem_size, 
     +     min_xyz, max_xyz, hkl_min, hkl_max, max_z_slab,
     +     first_rr_out, last_rr_out, max_rr_out_slab,
     +     first_u, last_u, max_u_slab,
     +     first_y, last_y, max_y_slab,
     +     interpolants_size
      
      integer nrffts_1_row, nrffts_1_col, nrffts_1_omega,
     +     nrffts_2_row, nrffts_2_col, nrffts_2_omega
      
      integer alloc_stat, ierr
            
C     The next array variables have their lengths defined dynamically
      real, allocatable :: f_vru(:), f_vru_tmp(:), F_XK_dZ(:),
     +     rho(:), interpolants(:), density_map(:), ctf_val(:),
     +     abs_ctf(:), disef(:),map3d(:), rffta_omega(:)
      
      complex, allocatable :: f_hkl(:),f_hkl_ref(:),fft_interp(:),
     +     fft3d(:),fft_klh(:)
      
      character(len=18) :: subname = 'PCTFR main routine'
      character(len=64) fname
      
      REAL    MAP_MIN, MAP_MAX, Weight
     +     , mapprjbig, imgprjbig, enlargefact
      real    angles_rad(1:3), rot_mat(1:3,1:3)
      integer fft_ann
      
      integer i, j, k 
      
C     *********************************************************************
C     *                      Initialization                               *
C     *********************************************************************
      
      progname = 'PCTFR'
      test_time   = .true.
      test_memory = .true.
      call usage
C     INITIALIZE MPI
C     **************
      call MPI_INIT ( mpi_err )
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_INIT')
      call MPI_COMM_RANK ( MPI_COMM_WORLD, iam, mpi_err )
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_COMM_RANK')
      call MPI_COMM_SIZE ( MPI_COMM_WORLD, num_nodes, mpi_err )
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_COMM_SIZE')
      if (num_nodes .gt. max_numnodes) then
         write(stdoutput, 9001) num_nodes, max_numnodes
 9001    format(//' Requested number of nodes = NUM_NODES    = ',i10/
     +        ' Available number of nodes = MAX_NUMNODES = ',i10/)
         call error_stop(progname, subname,  
     +        'NUM_NODES requested exceeds MAX_NUMNODES')
      endif

      t1 = wtime()

      if( iam .eq. 0 ) then
         print 1122, num_nodes
 1122    format(' Number of Nodes =',i5)

C     *********
C     Read User's Control-Input
C     I have change the control_input to multi files.

         call info

Cji   check the map, and get the col and row number
Cji   3D model should have the smallest pixel_size and largest box_size.
Cji   comapred with boxed images.

         call get_map_dimensions(in_3d_name, npix_col, npix_row, 
     $        angstrom_per_pix, ierr)
         if(ierr .gt. 0) then
            call error_stop('PCTFR', 'PCTFR',
     +           'Error in 3D map mode, it should be 20 or 40')
         endif

         print *, PIXSIZ_OUT, angstrom_per_pix, BOX_OUT, npix_row
         if ( npix_row .LT. BOX_OUT .or. 
     +        (angstrom_per_pix - PIXSIZ_OUT).gt.10E-5)
     +        call error_stop(progname, subname,
     +        '3D map should have smallest pixel and largest box_size')
         BOX_OUT    = npix_row
         PIXSIZ_OUT = angstrom_per_pix
      endif

      if( num_nodes .gt. 1 ) then
C     CALL BCAST_PARAMETERS
C     *******************************
C     Broadcast User's parameters and PIF File Header to other nodes.
         call bcast_parameters 
      endif

      view_perread = 1
C     CALL INTLZ_PARAMS
C     *****************
      call intlz_params(hkl_min, hkl_max, num_hkl, min_xyz, max_xyz, 
     +     nrffta_row, nrffta_col, nrffta_omega, nfc_2d_row, 
     +     nfc_2d_col, nrffts_1_row, nrffts_1_col, 
     +     nrffts_1_omega, nrffts_2_row, nrffts_2_col, 
     +     nrffts_2_omega, den_min_x, den_max_x, den_min_y, 
     +     den_max_y, den_min_z, den_max_z,
     +     confine_3D)
      
Cji 01/06/01 added 2 print statements after call of intlz_params
      if( iam .eq. 0 ) then
         print *, ' zero_fill after  call of INTLZ_PARAMS =',zero_fill
         print *, ' nrffta_row after call of INTLZ_PARAMS =',nrffta_row
      endif

      call intlz_arrays(hkl_max, min_xyz, max_xyz,
     +     min_z, max_z, max_views,
     +     first_hk, last_hk, max_hk_slab, first_iq, last_iq,
     +     max_iq_slab, first_rr_out, last_rr_out, max_rr_out_slab,
     +     first_u, last_u, max_u_slab,
     +     first_y, last_y, max_y_slab, first_z, last_z, max_z_slab,
     +     npix_row)

C ---------- End of Initializationl --- Synchronize all the nodes ------

      call mpi_barrier(mpi_comm_world, mpi_err)
      t2 = wtime()
      t_init = t2 - t1
      
C     **********************************************************************
C     *             3D DFT of input 3D density map                         *
C     **********************************************************************

      mem_size = npix_col * npix_row * max_y_slab
      allocate(map3d(mem_size), stat = alloc_stat)
      mem_map3d = dble(4 * mem_size) 
      if (alloc_stat .gt. 0) call error_stop (progname, subname,
     +     '3D density map memory allocation failed')
      mem_total = mem_total + mem_map3d
      mem_peak  = max(mem_peak, mem_total)
      
      call read_map(in_3d_name, map3d, npix_col, npix_row,
     +     max_ y_slab, mapprjbig)

Cji  3d DFT, and exchange 8 quater

      allocate(rffta_omega(nrffta_omega ), stat = alloc_stat)
      if(alloc_stat .gt. 0) call error_stop (progname, subname,
     +     'Error in allocating memory for nrffta_omega')

      call vrffti( nrffta_col, rffta_omega )

      mem_size = (hkl_max+1)* num_hkl * (last_y - first_y + 1) 
      allocate(fft3d(mem_size), stat = alloc_stat)
      mem_fft3d = dble(8 * mem_size) 
      if(alloc_stat .gt. 0) then
         call error_stop (progname, subname,
     +        '3D density fft map memory allocation failed')
      endif
      mem_total = mem_total + mem_fft3d
      mem_peak  = max(mem_peak, mem_total)

C     FFT3D, (X,Y,Z) real -> (H,K,Z) complex(0:Hmax, Kmin:Kmax, Z_slab
C     Zmin:Zmax) because I change intlz_params.f to make -hkl_min =
C     hkl_max = nfc_2d_row = nfc_2d_col

      do i = 0, last_y - first_y
         center(1) = (npix_row-1)/2.0
         center(2) = (npix_col-1)/2.0
         call fft_2dfft(map3d(i*npix_col*npix_row+1), 
     +        npix_row, npix_col, center,
     +        nrffta_omega, rffta_omega, nrffta_row, nrffta_col, 
     +        hkl_max, hkl_max, FFT3D(i*(hkl_max+1)*num_hkl+1))
      enddo

      deallocate(map3d)
      mem_total = mem_total - mem_map3d
      call mpi_barrier(mpi_comm_world, mpi_err)

C     need exchange data between nodes from Z_slab to H_slab

      if (num_nodes.gt.1) then
         mem_size = (hkl_max+1)*num_hkl*(last_y - first_y + 1)
         allocate(fft_klh(mem_size), stat = alloc_stat)
         if(alloc_stat .gt. 0) then
            call error_stop (progname, subname,
     +           '3D density fft exchange memory 1 allocation failed')
         endif
         mem_total = mem_total + 8*mem_size 
         mem_peak  = max(mem_peak, mem_total)
         
         call arrange_3d_1(FFT3D, fft_klh, hkl_min, hkl_max)
         
         deallocate(fft3d)
         mem_total = mem_total - mem_fft3d
         mem_fft3d = 8 * mem_size
         
         mem_size = (last_iq-first_iq+1)*num_hkl*npix_row
         allocate(fft3d(mem_size), stat = alloc_stat)
         if(alloc_stat .gt. 0) then
            call error_stop (progname, subname,
     +           '3D density fft exchange memory 2 allocation failed')
         endif
         mem_total = mem_total + 8*mem_size 
         mem_peak  = max(mem_peak, mem_total)
         
         call exch_3d_1(fft_klh, hkl_min, hkl_max, FFT3D, npix_row)
         
         deallocate(fft_klh)
         mem_total = mem_total - mem_fft3d
         mem_fft3d = 8 * mem_size
      endif

C     1D fft, (H,K,Z)-->(K,L,H) complex(Kmin:Kmax, Lmin:Lmax, H_slab 0:Hmax)

      mem_size = (last_iq-first_iq+1)* num_hkl**2
      allocate(fft_klh(mem_size), stat = alloc_stat)
      if(alloc_stat .gt. 0) then
         call error_stop (progname, subname,
     +        '3D density fft 1dfft allocation failed')
      endif
      mem_total = mem_total +8*mem_size 
      mem_peak  = max(mem_peak, mem_total)

      CALL cfft_1d(FFT3D, fft_klh, hkl_min, hkl_max, npix_row,
     +     nrffta_omega, rffta_omega)

      deallocate(fft3d)
      mem_total = mem_total - mem_fft3d
      mem_fft3d = 8 * mem_size

C     we apply -tempfac to the images, so does the map apply tmpfac to
C     the slab of map.

      if (ctfmode.GT.0 .and. ctfmode.LE.4) then
         call maptempfac(fft_klh, hkl_max, TEMPFAC, PIXSIZ_OUT, units,
     +        frst_lst_iq(1,iam), frst_lst_iq(2,iam))
      endif

C     gather 3Dfft fft_klh in to a full 3D fft map fft3d use up to
C     annulus to save memory.

      fft_ann = int(real(nrffta_row)*PIXSIZ_OUT/res_max) + 1
      mem_size = (2*fft_ann+1)*(2*fft_ann+1)*(fft_ann+1)
      allocate(fft3d(mem_size), stat = alloc_stat)
      if(alloc_stat .gt. 0) then
         call error_stop (progname, subname,
     +        '3D fft memory allocation failed')
      endif
      mem_total = mem_total + 8 * mem_size
      mem_peak  = max(mem_peak, mem_total)

      call gather3d(fft_klh, hkl_max, fft3d, fft_ann)
      
      deallocate(fft_klh)
      mem_total = mem_total - mem_fft3d
      mem_fft3d = dble(8 * mem_size) 

      call mpi_barrier(mpi_comm_world, mpi_err)
      t1 = wtime()
      t_3dfft = t1 - t2

C **********************************************************************
C *             refine CTF of all input files               *
C **********************************************************************
C
Cji  initialize the CTF mem for half fft_size matrix

      mem_size = (nfc_2d_row +1) * (nfc_2d_col * 2 +1)
      allocate(ctf_val(mem_size), abs_ctf(mem_size),
     +     disef(mem_size), stat = alloc_stat)
      mem_ctf_val = dble( 12 * mem_size)
      
      if(alloc_stat .gt. 0) then
         call error_stop (progname, subname,
     +        ' Memory allocation failed')
      endif
      
      mem_total = mem_total + mem_ctf_val
      mem_peak  = max(mem_peak, mem_total)

Cji  loop for all the input parameter files to do cmpt_ort.
      t_analy     = 0
      mem_cmpt_intrps = 0
      
      do i = 1, NFILES

Cji   initialize the parameters for one image file.
         npix_row         =  NBOX(i)
         npix_col         =  NBOX(i)
         max_views        =  NIDS(i)
         angstrom_per_pix = PIXSIZS(i)

         if (iam .eq. 0) then
            pix_file_name    = IMG_FNAMES(i)
            ort_file_name    = PARAM_FNAMES(i)
C     ******PIF_OPEN************
            call PIF_OPEN (pix_file_name, pix_file_unit, 'r')

            call get_1st_ortid(j, ort_file_name)
            call read_1_pif(j, imgprjbig)
            enlargefact = mapprjbig/(imgprjbig*sqrt(real(nrffta_col)))
            print *, 'enlargefact=',enlargefact, imgprjbig, mapprjbig
         endif

         call mpi_bcast(enlargefact, 1, mpi_real,
     +        0, mpi_comm_world, mpi_err)
         if( mpi_err .ne. 0 ) call error_stop(progname,subname,
     +        'Error in calling MPI_BCAST enlargefact')

Cji  initialize the CTF value  for half fft_size matrix
         do k = 1, (nfc_2d_row +1) * (nfc_2d_col * 2 +1)
            ctf_val(k) = 0.0
            abs_ctf(k) = 1.0
            disef(k) = 1.0
         enddo
         
         IF (funcweight .GT.0.0 ) then 
            call disef_para(disef, nfc_2d_row, nfc_2d_col, funcweight,
     +           angstrom_per_pix, units)
         endif

C     read image files, and refine its CTF
C     CALL CMPT_INTRPS
C     ****************
         mem_cmpt = 0

         call cmpt_ctf(fft3d, hkl_min, hkl_max, fft_ann,
     +        t_analy, mem_cmpt, ctf_val, abs_ctf
     +        , disef, i, enlargefact)
         
         mem_cmpt_intrps = max(mem_cmpt_intrps, mem_cmpt)
         mem_peak  = max(mem_peak, mem_total+mem_cmpt)

Cji    finish read the pif file, Close PIF file.

         if(  iam .eq. 0 ) then
            call pif_close (pix_file_unit)
            print *, 'Finished image file: ', pix_file_name
         endif
         
      enddo

      deallocate(ctf_val, abs_ctf, disef, fft3d)
      mem_total = mem_total - mem_ctf_val - mem_fft3d

Cji end of all the parameter files

      call mpi_barrier(mpi_comm_world, mpi_err)
      t2 = wtime()
      t_interp = t2 - t1 - t_analy

C ***************************************************************************
C
      if (mem_total .ne. 0) print *,'mem_total=',mem_total,
     +     'At the end, mem_total should be 0!!!'

      if ( iam .eq. 0 ) then
C     Print timing message.
         time = t_init + t_3dfft + t_analy + t_interp
         
         if (test_time) 
     +        print 40, iam, t_init,
     +        t_3dfft,
     +        t_analy,
     +        t_interp,
     +        time
 40      format (/ 'Node ', i3, 
     +        ' Time used in initialization                      =',
     +        f15.4, ' ' / 8x,
     +        ' Time used in read map & 3DFFT                    =',
     +        f15.4, ' ' / 8x,
     +        ' Time used in Read & FFT images & 2D Cut          =',
     +        f15.4, ' ' / 8x,
     +        ' Time used in refine CTF                          =', 
     +        f15.4, ' ' / 8x, 
     +        ' Total time used                                  =',
     +        f15.4)

C     Print memory allocating message.
         if (test_memory) print 60, iam, mem_peak, 
     +        mem_map3d, mem_fft3d, mem_cmpt_intrps
 60      format(/ 'Node ', i3, ' Memory allocation in BYTES:', /
     +        8x, ' Peak memory allocated   =', f18.1 /
     +        8x, ' Memory for 3D Map       =', f18.1 /
     +        8x, ' Memory for 3D FFT       =', f18.1 /
     +        8x, ' Memory used in CMPT_CTF =', f18.1 )

         print *, '**********finished nicely**********'
      endif

      call MPI_FINALIZE ( mpi_err )

      stop
      end program pctfr
